@@ -34,13 +34,12 @@ fastTypeOf tyenv expr =
     ERec fields   -> tRec [ (name, fastTypeOf tyenv e) | (name, e) <- fields ]
     ESel e sel    -> typeSelect (fastTypeOf tyenv e) sel
     EIf _ e _     -> fastTypeOf tyenv e
-    EComp t _ _   -> t
+    EComp len t _ _ -> tSeq len t
     EAbs x t e    -> tFun t (fastTypeOf (Map.insert x (Forall [] [] t) tyenv) e)
     EApp e _      -> case tIsFun (fastTypeOf tyenv e) of
                         Just (_, t) -> t
                         Nothing     -> panic "Cryptol.TypeCheck.TypeOf.fastTypeOf"
                                          [ "EApp with non-function operator" ]
-    ECast _ t     -> t
     -- Polymorphic fragment
     EVar      {}  -> polymorphic
     ETAbs     {}  -> polymorphic
@@ -93,16 +92,18 @@ fastSchemaOf tyenv expr =
     EComp  {}      -> monomorphic
     EApp   {}      -> monomorphic
     EAbs   {}      -> monomorphic
-    ECast  {}      -> monomorphic
   where
     monomorphic = Forall [] [] (fastTypeOf tyenv expr)
 
 -- | Yields the return type of the selector on the given argument type.
 typeSelect :: Type -> Selector -> Type
 typeSelect (TUser _ _ ty) sel = typeSelect ty sel
-typeSelect (TCon _tctuple ts) (TupleSel i _) = ts !! i
+typeSelect (tIsTuple -> Just ts) (TupleSel i _)
+  | i < length ts = ts !! i
 typeSelect (TRec fields) (RecordSel n _)
-     | Just ty <- lookup n fields = ty
-typeSelect (TCon _tcseq [_, a]) (ListSel _ _) = a
+  | Just ty <- lookup n fields = ty
+typeSelect (tIsSeq -> Just (_, a)) ListSel{} = a
+typeSelect (tIsSeq -> Just (n, a)) sel@TupleSel{} = tSeq n (typeSelect a sel)
+typeSelect (tIsSeq -> Just (n, a)) sel@RecordSel{} = tSeq n (typeSelect a sel)
 typeSelect ty _ = panic "Cryptol.TypeCheck.TypeOf.typeSelect"
                     [ "cannot apply selector to value of type", render (pp ty) ]
